#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <string.h>

#ifndef PORT
#define PORT 10000
#endif

#define CMDSZ  16
#define PINSZ   15
#define BUFSZ  (CMDSZ * 4)

char pin[PINSZ] = "148554952184403";

int connections;

void check_pin(const char *cmd)
{
	long i;

	for (i = 0; i < PINSZ; i++)
		if (cmd[i] != pin[i])
			break;
	if (i == PINSZ)
		printf("PIN OK\n");
}

void get_data(const char *cmd)
{
	long reply;

	switch (*cmd) {
	case 'p':
		reply = getpid();
		break;
	case 'n':
		reply = connections;
		break;
	}
	printf("%ld\n", reply);
}

void child()
{
	char cmd[BUFSZ + 1];
	char *end, *scan;
	size_t len;
	int n;

	for (;;) {
		if (fgets(cmd, BUFSZ + 1, stdin) == NULL)
			break;
		end = index(cmd, '\n');
		if (end == NULL) {
			fprintf(stderr, "no new lines in %d bytes, ignored\n", BUFSZ - 1);
			continue;
		}
		*end = '\0';
		scan = cmd;
		for (len = strlen(scan); len >= CMDSZ; len -= CMDSZ, scan += CMDSZ) {
			switch (*scan) {
			case 'p':
				check_pin(scan + 1);
				break;
			case 'g':
				get_data(scan + 1);
				break;
			default:
				fprintf(stderr, "unknown command: %c, ignored\n", *scan);
				break;
			}
		}
		if (len) {
			fprintf(stderr, "incomplete command '%s', ignored\n", scan);
		}
	}
}

int main()
{
	int lstn;
	int enable;
	struct sockaddr_in lstn_addr;

	lstn = socket(AF_INET, SOCK_STREAM, 0);
	if (lstn < 0) {
		perror("socket");
		return 1;
	}
	enable = 1;
	if (setsockopt(lstn, SOL_SOCKET, SO_REUSEADDR, &enable, sizeof(enable)) < 0) {
		perror("setsockopt");
		return 1;
	}
	bzero(&lstn_addr, sizeof(lstn_addr));

	lstn_addr.sin_family = AF_INET;
	lstn_addr.sin_addr.s_addr = htonl(INADDR_ANY);
	lstn_addr.sin_port = htons(PORT);

	if (bind(lstn, (struct sockaddr *)&lstn_addr, sizeof(lstn_addr)) < 0) {
		perror("bind");
		return 1;
	}

	if (listen(lstn, 10) < 0) {
		perror("listen");
		return 1;
	}
	printf("Listening on port %d\n", PORT);

	signal(SIGCHLD, SIG_IGN);

	for (;;) {
		int con = accept(lstn, NULL, NULL);
		if (con < 0) {
			perror("accept");
			return 1;
		}
		connections++;

		switch (fork()) {
		case -1:
			perror("fork");
			return 1;
		case 0:
			printf("New connection, child %d\n", getpid());

			close(0);
			dup(con);
			close(1);
			dup(con);
			close(2);
			dup(con);
			child();
			exit(0);
			break;
		default:
			close(con);
			break;
		}
	}
	return 0;
}
